/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.io.IOException;
import java.util.Iterator;
import java.util.UUID;
import java.util.concurrent.ConcurrentNavigableMap;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.mapdb.BTreeKeySerializer;
import org.mapdb.DB;
import org.mapdb.Fun.Pair;

import com.google.common.collect.Interner;
import com.google.common.collect.Interners;

/**
 * Implementation using MapDB (https://github.com/jankotek/mapdb) as back-end.
 * 
 * @author F. Mannhardt
 * 
 */
final class MapDBStore implements ExternalStore {

	private static final int KEY_POOL_LIMIT = 16384;
	private static final int LITERAL_POOL_LIMIT = KEY_POOL_LIMIT * 8;

	public static class Builder {

		private DB db;
		private String storeId;
		private Iterator<Pair<Long, XAttributeExternalImpl>> pumpSource;

		public Builder withDB(DB db) {
			this.db = db;
			return this;
		}

		public Builder withLog(String storeId) {
			this.storeId = storeId;
			return this;
		}

		public Builder withDataPump(Iterator<Pair<Long, XAttributeExternalImpl>> pumpSource) {
			this.pumpSource = pumpSource;
			return this;
		}

		public MapDBStore build() {
			if (db == null) {
				// Stand-alone mode with temporary file
				MapDBDatabase dbConnection = new MapDBDatabase();
				try {
					db = dbConnection.createDB();
				} catch (IOException e) {
					throw new RuntimeException("Failed to access MapDB database file!");
				}
			}
			if (storeId == null) {
				storeId = UUID.randomUUID().toString();
			}
			return new MapDBStore(this);
		}

	}

	private final String storeId;

	private final ConcurrentNavigableMap<Long, XAttributeExternalImpl> attributeStorage;

	private final DB db;
	private final IdFactory idFactory;
	private final StringPool keyPool;
	private final StringPool literalPool;
	private final Interner<String> interner;

	private boolean useCachingWrapper = true;

	private MapDBStore(Builder builder) {
		super();
		this.storeId = builder.storeId;
		this.idFactory = getIDFactory(builder.db);
		this.keyPool = getKeyPool(builder.db);
		this.literalPool = getLiteralPool(builder.db);
		this.db = builder.db;
		this.interner = Interners.newStrongInterner();

		if (db.exists(prefix(".attributeStore"))) {
			attributeStorage = db.treeMap(prefix(".attributeStore"));
		} else {
			if (builder.pumpSource != null) {
				attributeStorage = db.treeMapCreate(prefix(".attributeStore")).keySerializer(BTreeKeySerializer.LONG)
						.valueSerializer(new MapDBAttributeSerializer())
						// Trading random access speed for faster sequential access
						//.valuesOutsideNodesEnable()						
						.nodeSize(64).pumpSource(builder.pumpSource).makeOrGet();
			} else {
				attributeStorage = db.treeMapCreate(prefix(".attributeStore")).keySerializer(BTreeKeySerializer.LONG)
						.valueSerializer(new MapDBAttributeSerializer())
						// Trading random access speed for faster sequential access
						//.valuesOutsideNodesEnable()
						.nodeSize(64).makeOrGet();
			}
		}
	}

	private String prefix(String s) {
		return storeId.concat(s);
	}

	private StringPool getKeyPool(DB db) {
		StringPool pool = db.<StringPool>get(prefix(".stringPool"));
		if (pool != null) {
			return pool;
		} else {
			return new StringPoolCASImpl(KEY_POOL_LIMIT);
		}
	}

	private StringPool getLiteralPool(DB db) {
		StringPool pool = db.<StringPool>get(prefix(".literalPool"));
		if (pool != null) {
			return pool;
		} else {
			return new StringPoolCASImpl(LITERAL_POOL_LIMIT);
		}
	}

	private IdFactory getIDFactory(DB db) {
		return new MapDBSeqFactory(KEY_POOL_LIMIT, db.<Long>get(prefix(".seqNr")));
	}

	final ConcurrentNavigableMap<Long, XAttributeExternalImpl> getMapStorage() {
		return attributeStorage;
	}

	/**
	 * Returns the composite key of the attribute and the owner. Assumes that
	 * the internal key of the attribute is already added to the pool.
	 * 
	 * @param owner
	 * @param attributekey
	 * @return the composite key
	 */
	static Long getCompositeKey(ExternalAttributable owner, int attributekey) {
		return owner.getExternalId() + attributekey;
	}

	/**
	 * @param owner
	 * @param compositeKey
	 * @return
	 */
	static int getInternalAttributeKey(ExternalAttributable owner, long compositeKey) {
		return (int) (compositeKey - owner.getExternalId());
	}

	/**
	 * Returns the internal key of the attribute with the supplied
	 * {@link String} key. In case no such attribute is know, NULL is returned.
	 * 
	 * @param owner
	 * @param attributeKey
	 * @return the composite key or NULL is no such key exists
	 */
	final Integer getInternalAttributeKey(ExternalAttributable owner, String attributeKey) {
		return keyPool.getIndex(attributeKey);
	}

	/**
	 * @param owner
	 * @return
	 */
	final long lowestCompositeKey(ExternalAttributable owner) {
		return owner.getExternalId();
	}

	/**
	 * @param owner
	 * @return
	 */
	final long highestCompositeKey(ExternalAttributable owner) {
		return owner.getExternalId() + (getAttributeKeyPool().getPoolLimit() - 1);
	}

	private final XAttributeMap getAttributeMap(ExternalAttributable attributable) {
		XAttributeMap attributeMap = null;
		if (useCachingWrapper) {
			attributeMap = new MapDBAttributeMapWithCache(attributable, this, interner);
		} else {
			attributeMap = new MapDBAttributeMap(attributable, this);
		}
		return attributeMap;
	}

	public final StringPool getAttributeKeyPool() {
		return keyPool;
	}

	public StringPool getLiteralPool() {
		return literalPool;
	}

	public final IdFactory getIdFactory() {
		return idFactory;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#getAttributes
	 * (long)
	 */
	@Override
	public final XAttributeMap getAttributes(final ExternalAttributable attributable) {
		return getAttributeMap(attributable);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#hasAttributes
	 * (org.progressmining.xeslite.external.store.XExternalAttributable)
	 */
	@Override
	public final boolean hasAttributes(final ExternalAttributable attributable) {
		//TODO somehow cache this
		return !attributeStorage.keySet()
				.subSet(lowestCompositeKey(attributable), true, highestCompositeKey(attributable), true).isEmpty();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#setAttributes
	 * (long, org.deckfour.xes.model.XAttributeMap)
	 */
	@Override
	public final void setAttributes(final ExternalAttributable attributable, final XAttributeMap attributes) {

		if (attributes.isEmpty()) {
			getAttributeMap(attributable).clear();
			return;
		}

		XAttributeMap attributeMap = getAttributeMap(attributable);
		Iterator<XAttribute> values = attributeMap.values().iterator();
		while (values.hasNext()) {
			XAttribute a = values.next();
			if (!attributes.containsKey(a.getKey())) {
				values.remove();
			}
		}
		attributeMap.putAll(attributes);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#removeAttributes
	 * (long)
	 */
	@Override
	public final XAttributeMap removeAttributes(final ExternalAttributable attributable) {
		XAttributeMap attributeMap = getAttributeMap(attributable);

		XAttributeMap oldMap = new XAttributeMapImpl(attributeMap.size());
		// Use values() instead of putAll() to avoid two copies (1 by entrySet() that is called by putAll(), 2 by us)
		Iterator<XAttribute> values = attributeMap.values().iterator();
		while (values.hasNext()) {
			XAttribute a = values.next();
			oldMap.put(a.getKey(), a);
		}
		attributeMap.clear();
		return oldMap;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.progressmining.xeslite.external.store.XAttributeStore#commit()
	 */
	@Override
	public final void commit() {
		//no-op as we do not support transactions anyway
	}

}