/*
 * Copyright (c) 2014 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension.StandardModel;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.XVisitor;
import org.deckfour.xes.util.XAttributeUtils;
import org.processmining.xeslite.external.XEventObservableCachingExternalImpl.CachedEventObserver;

import com.google.common.primitives.Ints;
import com.google.common.primitives.Longs;

/**
 * Compressed storage of a trace in form of event identifies
 * 
 * @author F. Mannhardt
 * 
 */
final class XTraceCompressedExternalImpl extends XAbstractCompressedList<XEvent> implements XTrace,
		ExternalAttributable {

	private final class CachedEventObserverImpl implements CachedEventObserver {
		private final long id;
		private final int index;

		private CachedEventObserverImpl(long id, int index) {
			this.id = id;
			this.index = index;
		}

		public void setTransition(String transition) {
			EventData eventData = getEventData();
			ByteBuffer data = findData(index, id, eventData);
			if (data != null) {
				StandardModel standardTransition = StandardModel.decode(transition);
				storeTransitionExtension(standardTransition, data);
				setEventData(eventData);
			}
		}

		public void setTime(long time) {
			EventData eventData = getEventData();
			ByteBuffer data = findData(index, id, eventData);
			if (data != null) {
				storeTimeExtension(new Date(time), data);
				setEventData(eventData);
			}
		}

		public void setName(String name) {
			EventData eventData = getEventData();
			ByteBuffer data = findData(index, id, eventData);
			if (data != null) {
				storeConceptExtension(name, data);
				setEventData(eventData);
			}
		}

		private final ByteBuffer findData(int index, long id, EventData eventData) {
			long currentId = eventData.ids[index];
			if (currentId == id) {
				return eventData.data[index];
			} else {
				for (int i = 0; i < eventData.ids.length; i++) {
					currentId = eventData.ids[i];
					if (currentId == id) {
						return eventData.data[i];
					}
				}
			}
			return null;
		}

	}

	private static final int COMMON_ATTRIBUTES_SIZE = Longs.BYTES + Ints.BYTES + Ints.BYTES;
	private static final StandardModel[] LIFECYLCES = StandardModel.values();
	
	private final ExternalStore store;
	private final long id;

	XTraceCompressedExternalImpl(boolean cacheCommonAttributes, ExternalStore attributeStore) {
		this(cacheCommonAttributes, null, attributeStore);
	}

	XTraceCompressedExternalImpl(boolean cacheCommonAttributes, XAttributeMap attributeMap, ExternalStore attributeStore) {
		this(attributeStore.getIdFactory().nextId(), cacheCommonAttributes, attributeMap, attributeStore);
	}

	XTraceCompressedExternalImpl(long id, boolean cacheCommonAttributes, XAttributeMap attributeMap, ExternalStore attributeStore) {
		super(cacheCommonAttributes ? COMMON_ATTRIBUTES_SIZE : 0);
		this.store = attributeStore;
		this.id = id;
		if (attributeMap != null) {
			XAttributeMap newAttributeMap = getAttributes();
			for (XAttribute a : attributeMap.values()) {
				newAttributeMap.put(a.getKey(), a);
			}
		}
	}

	protected XEvent newInstance(final int index, final long id, ByteBuffer data) {

		if (hasPayload()) {

			XEventObservableCachingExternalImpl event = new XEventObservableCachingExternalImpl(id, null, store);

			if (data != null) {

				long timestamp;
				timestamp = data.getLong();
				if (timestamp != -1) {
					event.setTime(timestamp);
				}

				int lifecycle = data.getInt();
				if (lifecycle != -1) {
					event.setTransition(LIFECYLCES[lifecycle].getEncoding());
				}

				int nameIndex = data.getInt();
				if (nameIndex != -1) {
					event.setName(store.getLiteralPool().getValue(nameIndex));
				}

				assert !data.hasRemaining();
			}

			event.setObserver(new CachedEventObserverImpl(id, index));

			return event;
		} else {
			return new XEventBareExternalImpl(id, null, store);
		}
	}

	protected ByteBuffer convertData(XEvent e) {
		
		if (hasPayload()) {
			ByteBuffer data = ByteBuffer.allocate(COMMON_ATTRIBUTES_SIZE);

			storeTimeExtension(XTimeExtension.instance().extractTimestamp(e), data);
			storeTransitionExtension(XLifecycleExtension.instance().extractStandardTransition(e), data);
			storeConceptExtension(XConceptExtension.instance().extractName(e), data);
			
			return data;
			
		} else {
			return null;
		}
		
	}

	private void storeConceptExtension(String name, ByteBuffer data) {
		if (name != null) {
			Integer index = store.getLiteralPool().put(name);
			data.putInt(12, index);
		} else {
			data.putInt(12, -1);
		}
	}

	private static void storeTransitionExtension(StandardModel lifecycleTransition, ByteBuffer data) {
		if (lifecycleTransition != StandardModel.UNKNOWN && lifecycleTransition != null) {
			data.putInt(8, lifecycleTransition.ordinal());
		} else {
			data.putInt(8, -1);
		}
	}

	private void storeTimeExtension(Date time, ByteBuffer data) {
		if (time != null) {
			data.putLong(0, time.getTime());
		} else {
			data.putLong(0, -1l);
		}
	}

	protected XEvent convertElement(XEvent e) {
		XEventBareExternalImpl newEvent;
		if (!(e instanceof XEventBareExternalImpl)) {
			newEvent = new XEventCachingExternalImpl(e.getAttributes(), store);
		} else {
			newEvent = (XEventBareExternalImpl) e;
		}
		return newEvent;
	}

	protected long getExternalId(XEvent e) {
		assert e instanceof XEventBareExternalImpl : "Invalid event " + e + " should be of class "
				+ XEventBareExternalImpl.class.getSimpleName();
		return ((XEventBareExternalImpl) e).getExternalId();
	}
	
	public long getExternalId() {
		return id;
	}

	@Override
	public XAttributeMap getAttributes() {
		return store.getAttributes(this);
	}

	@Override
	public Set<XExtension> getExtensions() {
		return XAttributeUtils.extractExtensions(getAttributes());
	}

	@Override
	public void setAttributes(XAttributeMap attributes) {
		store.setAttributes(this, attributes);
	}

	@Override
	public boolean hasAttributes() {
		return store.hasAttributes(this);
	}

	@Override
	public Object clone() {
		XTrace clone = new XTraceCompressedExternalImpl(hasPayload(), getAttributes(), store);
		List<XEvent> clonedEvents = new ArrayList<>();
		for (XEvent event : this) {
			clonedEvents.add((XEvent) event.clone());
		}
		clone.addAll(clonedEvents);
		return clone;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.deckfour.xes.model.XTrace#insertOrdered(org.deckfour.xes.model.XEvent
	 * )
	 */
	public int insertOrdered(XEvent event) {
		// Same as in OpenXES
		if (this.size() == 0) {
			// append if list is empty
			add(event);
			return 0;
		}
		XAttribute insTsAttr = event.getAttributes().get(XTimeExtension.KEY_TIMESTAMP);
		if (insTsAttr == null) {
			// append if event has no timestamp
			add(event);
			return (size() - 1);
		}
		Date insTs = ((XAttributeTimestamp) insTsAttr).getValue();
		for (int i = (size() - 1); i >= 0; i--) {
			XAttribute refTsAttr = get(i).getAttributes().get(XTimeExtension.KEY_TIMESTAMP);
			if (refTsAttr == null) {
				// trace contains events w/o timestamps, append.
				add(event);
				return (size() - 1);
			}
			Date refTs = ((XAttributeTimestamp) refTsAttr).getValue();
			if (insTs.before(refTs) == false) {
				// insert position reached
				add(i + 1, event);
				return (i + 1);
			}
		}
		// beginning reached, insert at head
		add(0, event);
		return 0;
	}

	@Override
	public void accept(XVisitor visitor, XLog log) {
		// Same as in OpenXES
		visitor.visitTracePre(this, log);
		for (XAttribute attribute : getAttributes().values()) {
			attribute.accept(visitor, this);
		}
		for (XEvent event : this) {
			event.accept(visitor, this);
		}
		visitor.visitTracePost(this, log);
	}

}