package org.processmining.xeslite.query;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.processmining.xeslite.query.extension.NotEqual;
import org.processmining.xeslite.query.syntax.Node;
import org.processmining.xeslite.query.syntax.QueryAtom;
import org.processmining.xeslite.query.syntax.QueryBinaryExpression;
import org.processmining.xeslite.query.syntax.QueryComposite;
import org.processmining.xeslite.query.syntax.QueryParserVisitor;
import org.processmining.xeslite.query.syntax.QueryRoot;
import org.processmining.xeslite.query.syntax.SimpleNode;

import com.googlecode.cqengine.attribute.Attribute;
import com.googlecode.cqengine.query.Query;
import com.googlecode.cqengine.query.QueryFactory;

final class QueryBuilderVisitor<A extends XAttributable> implements QueryParserVisitor {

	private static final DateFormat DATE_FORMATTER = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

	private final AttributeTypeResolver attributeTypeResolver;

	private final XIndexedAttributable<A> indexed;

	public QueryBuilderVisitor(XIndexedAttributable<A> indexed, AttributeTypeResolver attributeTypeResolver) {
		this.indexed = indexed;
		this.attributeTypeResolver = attributeTypeResolver;
	}

	public Object visit(SimpleNode node, Object data) {
		throw new UnsupportedOperationException("There should be no unamed node");
	}

	public Object visit(QueryRoot node, Object data) {
		assert node.jjtGetNumChildren() == 1 : "Invalid root!";
		return node.jjtGetChild(0).jjtAccept(this, data);
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	public Object visit(QueryComposite node, Object data) {
		if (node.jjtGetNumChildren() == 2) {
			Node left = node.jjtGetChild(0);
			Node right = node.jjtGetChild(1);
			switch (node.op.toLowerCase()) {
				case "and" :
					return QueryFactory.and((Query) left.jjtAccept(this, data), (Query) right.jjtAccept(this, data));

				case "or" :
					return QueryFactory.or((Query) left.jjtAccept(this, data), (Query) right.jjtAccept(this, data));
			}
		}
		return node.jjtGetChild(0).jjtAccept(this, data);
	}

	public Object visit(QueryAtom node, Object data) {
		return node.val;
	}

	public Object visit(QueryBinaryExpression node, Object data) {
		assert node.jjtGetNumChildren() == 2 : "Invalid binary expression!";

		final Node left = node.jjtGetChild(0);
		final Node right = node.jjtGetChild(1);
		final String attributeName = ((String) left.jjtAccept(this, data)).trim();
		final String value = ((String) right.jjtAccept(this, data)).trim();

		switch (node.op.toLowerCase()) {
			case "=" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return QueryFactory.equal(attr, value);
					}

				});

			case "!=" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return new NotEqual<>(attr, value);
					}

				});

			case ">" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return QueryFactory.greaterThan(attr, value);
					}

				});

			case "<" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return QueryFactory.lessThan(attr, value);
					}

				});

			case "<=" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return QueryFactory.lessThanOrEqualTo(attr, value);
					}

				});

			case ">=" :
				return createQuery(attributeName, value, new QueryBuilder<A>() {

					public <T extends Comparable<T>> Query<A> build(Attribute<A, T> attr, T value) {
						return QueryFactory.greaterThanOrEqualTo(attr, value);
					}

				});

			case "%" :
				// Only for literals
				return QueryFactory.contains(indexed.stringAttribute(attributeName), value);

			case "~" :
				// Only for literals
				return QueryFactory.matchesRegex(indexed.stringAttribute(attributeName), value);

			default :
				throw new UnsupportedOperationException("Unsupported operator " + node.op);
		}
	}

	private final Query<A> createQuery(String attributeName, String value, QueryBuilder<A> queryBuilder) {
		Class<?> type = attributeTypeResolver.getAttributeType(attributeName);
		try {
			if ((type == XAttributeContinuous.class || type == XAttributeDiscrete.class)) {
				Double doubleValue = Double.parseDouble(value);
				return queryBuilder.build(indexed.doubleAttribute(attributeName), doubleValue);
			} else if (type == XAttributeTimestamp.class) {
				Date dateValue = DATE_FORMATTER.parse(value);
				return queryBuilder.build(indexed.dateAttribute(attributeName), dateValue);
			}
		} catch (NumberFormatException | java.text.ParseException e) {
			// fall through with string
		}
		return queryBuilder.build(indexed.stringAttribute(attributeName), value);
	}

}